from datetime import datetime, timezone
from uuid import RFC_4122, UUID

import pytest

from hive.chat import ChatMessage, tell_user, tell_user_errors
from hive.common import parse_datetime


def test_basic_operation(mock_messagebus, mock_channel):
    tell_user("bonjour!", channel=mock_channel)

    assert len(mock_messagebus.published_events) == 1
    event = mock_messagebus.published_events[0]
    assert event.routing_key == "chat.messages"
    message = event.message

    assert message.keys() == {"text", "sender", "timestamp", "uuid"}
    assert message["text"] == "bonjour!"
    assert message["sender"] == "hive"

    timestamp = parse_datetime(message["timestamp"])
    delta = (datetime.now(tz=timezone.utc) - timestamp).total_seconds()
    assert 0 <= delta < 1

    uuid = UUID(message["uuid"])
    assert uuid.variant == RFC_4122
    assert uuid.version == 4


def test_channel_creation(mock_messagebus):
    tell_user(ChatMessage(
        text="salop!",
        timestamp="2024-11-23 10:52:19.542344Z",
        uuid="urn:uuid:0669fa00-93d8-4c35-bccc-469258b9b065",
    ))

    assert len(mock_messagebus.published_events) == 1
    event = mock_messagebus.published_events[0]
    assert event.routing_key == "chat.messages"
    assert event.message == {
        "text": "salop!",
        "sender": "hive",
        "timestamp": "2024-11-23 10:52:19.542344+00:00",
        "uuid": "0669fa00-93d8-4c35-bccc-469258b9b065",
    }


def test_tell_user_errors(mock_messagebus):
    class TestError(Exception):
        pass

    with pytest.raises(TestError):
        with tell_user_errors():
            raise TestError("oh <no>!")

    assert len(mock_messagebus.published_events) == 1
    event = mock_messagebus.published_events[0]
    assert event.routing_key == "chat.messages"
    assert event.message.keys() == {
        "text", "html", "sender", "timestamp", "uuid",
    }
    assert event.message["text"] == "TestError: oh <no>!"
    assert event.message["html"].endswith(
        " <code>TestError: oh &lt;no&gt;!</code>",
    )



from __future__ import annotations

from datetime import datetime, timezone
from functools import cached_property
from typing import Any, Optional


class ClientEvent:
    """An undecorated `ClientEvent`, as reported by Matrix Commander.

    **7.2 Room event format**

    `ClientEvent` -- The format used for events returned from a
        homeserver to a client via the Client- Server API, or sent
        to an Application Service via the Application Services API.

    https://spec.matrix.org/v1.12/client-server-api/#room-event-format
    """
    def __init__(self, serialized: dict[str, Any]):
        self._event = serialized

    def __eq__(self, other):
        if not isinstance(other, ClientEvent):
            return False
        return self._event == other._event

    def json(self):
        return self._event

    @cached_property
    def content(self) -> EventContent:
        """The body of this event, as created by the client which sent it.
        """
        return EventContent(self._event["content"])

    @cached_property
    def event_id(self) -> str:
        """The globally unique identifier for this event.
        """
        return self._event["event_id"]

    @cached_property
    def event_type(self) -> str:
        """The type of the event, e.g. "m.room.message".
        """
        return self._event["type"]

    @cached_property
    def room_id(self) -> str:
        """The ID of the room associated with this event.
        """
        return self._event["room_id"]

    @cached_property
    def sender(self) -> str:
        """The fully-qualified ID of the user who sent this event.
        """
        return self._event["sender"]

    @cached_property
    def timestamp(self) -> datetime:
        """The timestamp on the originating homeserver when this
        event was sent.
        """
        return datetime.fromtimestamp(
            self._event["origin_server_ts"] / 1000,
            tz=timezone.utc,
        )


class EventContent:
    """The content of an "m.room.message" event.

    https://spec.matrix.org/v1.12/client-server-api/#mtext
    """
    def __init__(self, serialized: dict[str, Any]):
        self._content = serialized

    @cached_property
    def msgtype(self) -> str:
        """The type of the message, e.g. "m.text", "m.image".
        """
        return self._content["msgtype"]

    @cached_property
    def body(self) -> str:
        """The body of the message.
        """
        return self._content["body"]

    @cached_property
    def format(self) -> Optional[str]:
        """The format used in the formatted_body.
        """
        return self._content.get("format")

    @cached_property
    def formatted_body(self) -> str:
        """The formatted version of the `body`.
        """
        return self._content["formatted_body"]



import json

from hive.chat import ChatMessage
from hive.common import read_resource, parse_datetime


def read_event(filename):
    return json.loads(read_resource(filename))["source"]


def test_basic():
    event = read_event("resources/text.json")
    message = ChatMessage.from_matrix_event(event)
    assert message.text == "hello world"
    assert message.html is None
    assert message.sender == "user"
    assert message.timestamp == parse_datetime("2024-10-30 00:26:28.020Z")
    assert message.in_reply_to is None
    assert message.matrix.json() == event
    assert not message.has_unhandled_fields
    assert ChatMessage.from_json(message.json()) == message


def test_html():
    event = read_event("resources/html.json")
    message = ChatMessage.from_matrix_event(event)
    assert message.text == "hello **WORLD**"
    assert message.html == "hello <strong>WORLD</strong>"
    assert message.sender == "user"
    assert message.timestamp == parse_datetime("2024-11-26 00:14:10.740Z")
    assert message.in_reply_to is None
    assert message.matrix.json() == event
    assert not message.has_unhandled_fields
    assert ChatMessage.from_json(message.json()) == message


def test_image():
    event = read_event("resources/image.json")
    message = ChatMessage.from_matrix_event(event)
    assert message.text == "Wi-Fi_QR_code_Guest.jpg"
    assert message.html is None
    assert message.sender == "user"
    assert message.timestamp == parse_datetime("2024-10-26 22:49:53.880Z")
    assert message.in_reply_to is None
    assert message.matrix.json() == event
    assert not message.has_unhandled_fields
    assert ChatMessage.from_json(message.json()) == message



from __future__ import annotations

import logging

from dataclasses import dataclass
from html import escape
from threading import current_thread, main_thread
from typing import Optional
from uuid import UUID

from hive.messaging import Channel, blocking_connection

from .message import ChatMessage

logger = logging.getLogger(__name__)
d = logger.info


def tell_user(
        text: str | ChatMessage,
        *,
        channel: Optional[Channel] = None,
        **kwargs
) -> ChatMessage:
    """Send a ChatMessage.
    """
    if isinstance(text, ChatMessage):
        if kwargs:
            raise ValueError
        message = text
    else:
        message = ChatMessage(text=text, **kwargs)

    if channel:
        return _tell_user(channel, message)

    with blocking_connection(connection_attempts=1) as conn:
        return _tell_user(conn.channel(), message)


def _tell_user(channel: Channel, message: ChatMessage) -> ChatMessage:
    channel.publish_event(
        message=message.json(),
        routing_key="chat.messages",
    )
    return message


@dataclass
class tell_user_errors:
    channel: Optional[Channel] = None
    in_reply_to: Optional[str | UUID | ChatMessage] = None

    def __enter__(self) -> tell_user_errors:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not exc_type:
            return

        message_prefix = self._message_prefix

        logger.exception(f"{message_prefix}EXCEPTION")

        message = f"{message_prefix}{exc_type.__name__}: {exc_val}"

        tell_user(
            text=message,
            html=self._message_as_html(message),
            in_reply_to=self.in_reply_to,
            channel=self.channel,
        )

    @property
    def _message_prefix(self) -> str:
        if (thread := current_thread()) is main_thread():
            return ""
        if not (thread_name := thread.name):
            return ""
        return f"{thread_name}: "

    @staticmethod
    def _message_as_html(text: str) -> str:
        return (
            f'<strong style="background-color: red; color: yellow;">⚠</strong>'
            f" <code>{escape(text)}</code>"
        )



from __future__ import annotations

import json

from dataclasses import dataclass, field, fields
from datetime import datetime, timezone
from itertools import starmap
from types import NoneType
from typing import Any, Optional
from uuid import UUID, uuid4

from hive.common import parse_datetime, parse_uuid

from .matrix import ClientEvent as MatrixEvent


@dataclass
class ChatMessage:
    text: Optional[str] = None
    html: Optional[str] = None
    sender: str = "hive"
    timestamp: str | datetime = field(
        default_factory=lambda: datetime.now(tz=timezone.utc))
    uuid: str | UUID = field(default_factory=uuid4)
    in_reply_to: Optional[str | UUID | ChatMessage] = None
    matrix: Optional[dict | MatrixEvent] = None
    _unhandled: Optional[dict[str, Any]] = field(default=None, repr=False)

    def __post_init__(self):
        if not self.text and not self.html:
            raise ValueError
        if not isinstance(self.text, (str, NoneType)):
            raise TypeError(type(self.text))
        if not isinstance(self.html, (str, NoneType)):
            raise TypeError(type(self.html))

        if not isinstance(self.sender, str):
            raise TypeError(type(self.sender))

        if not isinstance(self.timestamp, datetime):
            self.timestamp = parse_datetime(self.timestamp)

        self.uuid = parse_uuid(self.uuid)

        if self.in_reply_to is not None:
            if isinstance(self.in_reply_to, ChatMessage):
                self.in_reply_to = self.in_reply_to.uuid
            self.in_reply_to = parse_uuid(self.in_reply_to)
            if self.in_reply_to == self.uuid:
                raise ValueError

        if not isinstance(self.matrix, (MatrixEvent, NoneType)):
            self.matrix = MatrixEvent(self.matrix)

    @classmethod
    def json_keys(cls) -> list[str]:
        names = (field.name for field in fields(cls))
        return [name for name in names if name[0] != "_"]

    @classmethod
    def from_json(cls, message: dict[str, Any]) -> ChatMessage:
        """Ultra-strict from-(deserialized)-JSON constructor.
        """
        if type(message) is not dict:
            raise TypeError
        if any(type(key) is not str for key in message.keys()):
            raise TypeError
        if type(message["sender"]) is not str:
            raise TypeError

        unhandled = message.copy()
        keys = cls.json_keys()
        values = [unhandled.pop(key, None) for key in keys]
        kwargs = dict(
            item
            for item in zip(keys, values)
            if item[1] not in ("", None)
        )

        if any(key != "matrix" and type(value) is not str
               for key, value in kwargs.items()):
            raise TypeError

        if unhandled:
            kwargs["_unhandled"] = unhandled

        return cls(**kwargs)

    @classmethod
    def from_matrix_event(
            cls,
            event: MatrixEvent | dict[str, Any],
    ) -> ChatMessage:
        if not isinstance(event, MatrixEvent):
            event = MatrixEvent(event)

        if event.event_type != "m.room.message":
            raise ValueError(event.event_type)

        content = event.content
        if content.msgtype not in {"m.text", "m.image"}:
            raise ValueError(content.msgtype)

        kwargs = {
            "text": content.body,
            "sender": "hive" if event.sender.startswith("@hive") else "user",
            "timestamp": event.timestamp,
            "matrix": event,
        }

        if content.format == "org.matrix.custom.html":
            kwargs["html"] = content.formatted_body

        return cls(**kwargs)

    @property
    def has_unhandled_fields(self):
        return bool(self._unhandled)

    def json(self) -> dict[str, Any]:
        items = ((key, getattr(self, key)) for key in self.json_keys())
        items = starmap(self._for_json, items)
        return dict(item for item in items if item)

    def _for_json(self, key, value) -> Optional[tuple[str, Any]]:
        if callable(func := getattr(value, "json", None)):
            value = func()
        if not value and value in (None, "", {}):
            return None  # omit
        try:
            _ = json.dumps(value)
            return key, value
        except Exception:
            return key, str(value)
